
/*
 *  N E W C R A T E R
 *
 *
 *
 *  Written by PM 2Ring 2001
 *
 *  Mar  7 2001
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

typedef unsigned long ULONG;
typedef unsigned char UBYTE;
typedef unsigned short USHORT;

typedef short int PIXEL;

extern ULONG intsqrt(ULONG n);

/* Wall slope parameters, in percentage form (fixed point) */
#define FIX 100
#define CWIDTH 60
#define CPEAK 90
#define REPORT 100000

long width=256, height=256;             /* Display size */
int cwidth = CWIDTH;
int cpeak = CPEAK;

PIXEL *ibuff = NULL;
UBYTE *obuff = NULL;

/* Plot a crater at (x,y) radius r */
void crater(int x, int y, int r);
void adjustrange(PIXEL *buff, long buffsize);
void cleanup(int rc);
void wall(int r);

/* Print error message & exit. */
void bail(char *s){if(s) fprintf(stderr, "%s!\n", s); cleanup(10);}

/* Pixel access macros */
#define pixel(x,y) (ibuff + (width*(y) + (x)))
#define pix(x,y) pixel(((x)+width)%width, ((y)+height)%height)

/* Plot */
#define plot(x,y,c) do{*pix(x,y) = (c);}while(0)

/* Centred Plot. Used for wall profile */
#define cplot(x,y,c) plot((x)+width/2,(y)+height/2,c)

/* Increment Plot */
#define iplot(x,y,c) do{*pix(x,y) += (c);}while(0)

/* Weighted Plot */
#define wplot(x,y,c) do{PIXEL *q=pix(x,y); int cc = c + (rim * (*q - c))/orad; *q = cc;}while(0)
//#define wplot(x,y,c) plot(x,y,c) 
                 
#define random(x) (((x)*(rand()>>15UL))>>16UL)

int main(int argc, char *argv[])
{
    long pixels, num=50000, seed=1, mr2;
    int i, j, rad, maxrad=0, rawf=0;
    PIXEL *p;

    /* Handle args */
    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            /* Skip rest of chars in option keyword */
            for (j=1; isalpha(argv[i][j]); j++)
            {}

            switch (argv[i][1])
            {
            case '?':
                fprintf(stderr, "Create a crater field in PGM.\nUsage:\n%s "
                 "[-widthN] [-heightN] [-numberN] [-maxradN] "
                 "[-groundN] [-topN] [-seedN] [-raw]\n"
                 , *argv);
                return 0;
                break;
            case 'g':
                cwidth = atoi(argv[i]+j);
                break;
            case 't':
                cpeak = atoi(argv[i]+j);
                break;
            case 'w':
                width = atoi(argv[i]+j);
                break;
            case 'h':
                height = atoi(argv[i]+j);
                break;
            case 'n':
                num = atoi(argv[i]+j);
                break;
            case 'm':
                maxrad = atoi(argv[i]+j);
                break;
            case 'r':
                rawf = 1;
                break;
            case 's':
                seed = atoi(argv[i]+j);
                break;
            default:
                break;
            }
        }
    }

    /* Allocate line buffer */
    pixels = width * height;
    if (!(ibuff = calloc(pixels, sizeof(PIXEL))))
        bail("No memory for data buffer");

    if (!(obuff = malloc(width)))
        bail("No memory for output buffer");

    if (maxrad <= 2)
        maxrad = width / 4;
    maxrad--;
    mr2 = maxrad * maxrad - 1;

    srand(seed);
    (void) rand();  /* Chuck away initial 1 */

    /* Process... */
    switch(num)
    {
    case 0:
        wall(maxrad);
        break;
    case 1:
        crater(width/2, height/2, maxrad);
        adjustrange(ibuff, pixels);
        break;
    default:
        for (i=1; i<=num; i++)
        {
            rad = 1 + maxrad / intsqrt(1 + random(mr2));
            /* fprintf(stderr, "%d\t", rad); */
            crater(random(width), random(height), rad);

            if (i%REPORT == 0)
              fprintf(stderr, "NewCrater: %d craters = %d%%\n", i, 100*i/num);
        }
        adjustrange(ibuff, pixels);
        break;
    }

    fprintf(stderr, "NewCrater: Saving image...\n");

    /* Send data to stdout */
    if (!rawf)  /* PGM raw header */
        printf("P5\n%d %d\n255\n", width, height);

    p = ibuff;
    for (i=0; i<height; i++)
    {
        int j;
        /* Convert 16 bit data to 8 bit */
        for (j=0; j<width; j++)
            obuff[j] = *p++;

        if (fwrite(obuff, 1, width, stdout) < width)
            bail("Error writing file data to stream");
    }

    cleanup(0);
    return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * */

/* Test crater wall algorithm & plot vertical cross-section */
void wall(int r)
{
    long b2;
    int j, c, b, d, w, h, irad, y;

    irad = (r * FIX) / (2*FIX - cwidth);
    w = (irad * cwidth) / FIX;
    b = (irad * FIX) / cpeak;
    b2 = b * b;
    h = intsqrt(b2 - w*w);

    for (d=0; d<width; d++)
        cplot(d,0, 128);

    for (j=0; j<=r; j++)
    {
        d = j<=irad ? j : 2*irad - j;
        y = intsqrt(b2 - d*d) - h;
        c = y>0 ? 160 : 255;
        cplot(-j, y, c);
        cplot(j, y, c);
    }
}

/* Plot a crater at (x,y) radius r */
void craterOLD(int x, int y, int r)
{
    long r2=r*r, b2, d2;
    int i, j, mj, c, b, d, w, h, irad;

    irad = (r * FIX) / (2*FIX - cwidth);
    w = (irad * cwidth) / FIX;
    b = (irad * FIX) / cpeak;
    b2 = b * b;
    h = intsqrt(b2 - w*w);

    /* Find average height around impact area. RR=reduced radius */
#if 0
  #define RR (r/2)
#else
  #define RR r
#endif    
    c = d = 0;
    for (i=-RR; i<=RR; i++)
        for (j=-RR; j<=RR; j++, c++)
            d += *pix(x+j,y+i);
    h += d / c;

    /* Plot */
    for (i=0; i<=r; i++)
    {
        mj = intsqrt(r2 - i*i);
        for (j=0; j<=mj; j++)
        {
            if (r < 3)  /* Handle small craters */
                c = h + random(5) - 2;
            else
            {
                /* Find depth (or height) & hence colour */
                d2 = i*i + j*j;
                d = intsqrt(d2);
                if (d>irad)
                {
                    d = 2*irad - d;
                    d2 = d * d;
                }
                if(d2 >= b2)   /* Kludge for big cpeak */
                    c = h;
                else
                    c = h - intsqrt(b2 - d2);
            }

            /* Use 4 way symmetry */
            plot(x+j, y+i, c);
            if (j) plot(x-j, y+i, c);
            if (i) plot(x+j, y-i, c);
            if (i && j) plot(x-j, y-i, c);
        }
    }
}

/* Plot a crater at (x,y) radius r */
void crater(int x, int y, int r)
{
    long r2=r*r, b2, d2, rim;
    int i, j, mj, c, b, d, w, h, irad, orad;

    irad = (r * FIX) / (2*FIX - cwidth);
    orad = r - irad;
    w = (irad * cwidth) / FIX;
    b = (irad * FIX) / cpeak;
    b2 = b * b;
    h = intsqrt(b2 - w*w);

    /* Find average height around impact area. RR=reduced radius */
#if 0
  #define RR irad //(r/2)
#else
  #define RR r
#endif    
    c = d = 0;
    for (i=-RR; i<=RR; i++)
        for (j=-RR; j<=RR; j++, c++)
            d += *pix(x+j,y+i);
    h += d / c;

    /* Plot */
    for (i=0; i<=r; i++)
    {
        mj = intsqrt(r2 - i*i);
        for (j=0; j<=mj; j++)
        {
            if (r < 3)  /* Handle small craters */
                c = h + random(5) - 2;
            else
            {
                /* Find depth (or height) & hence colour */
                d2 = i*i + j*j;
                d = intsqrt(d2);
                rim = d - irad;
                if (rim > 0)
                {
                    int dd = 2*irad - d;
                    d2 = dd * dd;
                }
                  
                if(d2 >= b2)   /* Kludge for big cpeak */
                    c = h;
                else
                    c = h - intsqrt(b2 - d2);
            }

            if(rim < 0)
            {
              /* Use 4 way symmetry */
              plot(x+j, y+i, c);
              if (j) plot(x-j, y+i, c);
              if (i) plot(x+j, y-i, c);
              if (i && j) plot(x-j, y-i, c);
            }
            else   /* Blend outer part into existing terrain using weighted mean */  
            {
              wplot(x+j, y+i, c);
              wplot(x-j, y+i, c);
              wplot(x+j, y-i, c);
              wplot(x-j, y-i, c);        
            }            
        }
    }
}

/* Adjust Dynamic Range */
void adjustrange(PIXEL *buff, long buffsize)
{
    int i;
    long lo=SHRT_MAX, hi=SHRT_MIN;

    /* Find limits */
    for (i=0; i<buffsize; i++)
    {
        if (buff[i]<lo)
            lo = buff[i];
        else if (buff[i]>hi)
            hi = buff[i];
    }
    fprintf(stderr, "NewCrater: lo=%d hi=%d\n", lo, hi);
    hi -= lo;
    if (hi <= 0)
        bail("No image");

    /* Rescale */
    for (i=0; i<buffsize; i++)
        buff[i] = (255 * (buff[i] - lo)) / hi;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * */

void cleanup(int rc)
{
    if (ibuff)
        free(ibuff);
    if (obuff)
        free(obuff);

    exit(rc);
}
